package filler

import (
	"fmt"
	"strings"

	"githouse.ru/macrox/gotygen/pkg/strcase"

	"githouse.ru/macrox/gotygen/pkg/utils"

	"githouse.ru/macrox/gotygen/pkg/token"

	"githouse.ru/macrox/gotygen/pkg/document"
	"githouse.ru/macrox/gotygen/pkg/scanner"
)

type jsonFiller struct {
	scanner.Scanner

	camelCtx strcase.CamelContext

	replacements map[string]string
	declaredName string
}

func (f *jsonFiller) deduceStruct(types []document.Type) document.Struct {
	result, m := document.NewStruct(), make(map[string][]*document.Field)
	var keys []string
	for _, t := range types {
		s := t.(document.Struct)
		for _, key := range s.Keys() {
			if _, ok := m[key]; !ok {
				m[key] = make([]*document.Field, 0, len(types))
				keys = append(keys, key)
			}
			if field, ok := s.Get(key); ok && field != nil {
				m[key] = append(m[key], field)
			}
		}
	}
	for _, key := range keys {
		fields := m[key]
		field := fields[0]
		field.Type = f.deduce(utils.SliceTransform(fields, func(field *document.Field, _ int) document.Type {
			return field.Type
		}))
		if fields[0].OmitEmpty = len(fields) != len(types); field.OmitEmpty {
			field.Type = field.Type.Nullable()
		}
		result.Set(field)
	}
	return result
}

func (f *jsonFiller) deduceMap(types []document.Type) document.Map {
	key := f.deduce(utils.SliceTransform(types, func(item document.Type, _ int) document.Type {
		return item.(document.Map).Key()
	}))
	value := f.deduce(utils.SliceTransform(types, func(item document.Type, _ int) document.Type {
		return item.(document.Map).Value()
	}))
	return document.NewMap(key, value)
}

func (f *jsonFiller) deduce(types []document.Type) document.Type {
	n := len(types)
	types = utils.RemoveType[document.Any](types)
	nullable := len(types) < n

	// Check types count
	switch len(types) {
	case 0:
		return document.NewAny()
	case 1:
		if nullable {
			return types[0].Nullable()
		} else {
			return types[0]
		}
	}

	// Check complex
	switch types[0].(type) {
	case document.String:
		if utils.AllIsType[document.String](types...) {
			return document.NewString().Nullable(nullable)
		}
	case document.Int:
		if utils.AllIsType[document.Int](types...) {
			return document.NewInt().Nullable(nullable)
		} else {
			if everyOk := utils.SliceEveryBy(types, func(item document.Type) bool {
				return utils.IsType[document.Int](item) || utils.IsType[document.Float](item)
			}); everyOk {
				return document.NewFloat().Nullable(nullable)
			}
		}
	case document.Float:
		if utils.AllIsType[document.Float](types...) {
			return document.NewFloat().Nullable(nullable)
		}
	case document.Bool:
		if utils.AllIsType[document.Bool](types...) {
			return document.NewBool().Nullable(nullable)
		}
	case document.Array:
		if utils.AllIsType[document.Array](types...) {
			for i := range types {
				types[i] = types[i].(document.Array).Element()
			}
			return document.NewArray(f.deduce(types))
		}
	case document.Struct:
		if utils.AllIsType[document.Struct](types...) {
			return f.deduceStruct(types).Nullable(nullable)
		} else {
			everyOk := utils.SliceEveryBy(types, func(item document.Type) bool {
				return utils.IsType[document.Struct](item) || utils.IsType[document.Map](item)
			})
			if everyOk {
				for i := range types {
					if s, ok := types[i].(document.Struct); ok {
						types[i] = document.NewMap(document.NewString(), f.deduce(utils.SliceTransform(s.Fields(), func(field *document.Field, _ int) document.Type {
							return field.Type
						})))
					}
				}
				return f.deduceMap(types)
			}
		}
	case document.Map:
		if utils.AllIsType[document.Map](types...) {
			return f.deduceMap(types)
		}
	}
	return document.NewAny()
}

func (f *jsonFiller) scanObjectType(refOut *document.Document, rootName string) (document.Type, error) {
	keys, types := make([]string, 0), make([]document.Type, 0)
	for f.More() {
		var (
			t   document.Type
			key string
			err error
		)
		_, key, err = f.Scan()
		if err != nil {
			return nil, err
		}
		keys = append(keys, key)
		if t, err = f.scanType(refOut, strcase.ToCamel(key, f.camelCtx)); err != nil {
			return nil, err
		} else {
			types = append(types, t)
		}
	}
	// Scan end token "}"
	if _, _, err := f.Scan(); err != nil {
		return nil, err
	}
	if len(keys) == 0 {
		return document.NewMap(document.NewString(), document.NewAny()), nil
	}
	if !validNames(keys) {
		var keyType document.Type
		if validIntegers(keys) {
			keyType = document.NewInt()
		} else {
			keyType = document.NewString()
		}
		return document.NewMap(keyType, f.deduce(types)), nil
	}
	s := document.NewStruct()
	for i, key := range keys {
		name, omitEmpty := strcase.ToCamel(key, f.camelCtx), utils.IsType[document.Map](types[i]) || utils.IsType[document.Array](types[i])
		if refOut != nil {
			if arr, ok := types[i].(document.Array); ok {
				element := arr.Element()
				if _, ok = element.(document.Struct); ok {
					id := f.normalizeId(rootName, name)
					refOut.PutDeclaredType(document.NewDeclaredType(id, element), true)
					arr.Element(document.NewRef(id))
				}
			} else if _, ok = types[i].(document.Struct); ok {
				id := f.normalizeId(rootName, name)
				refOut.PutDeclaredType(document.NewDeclaredType(id, types[i]), true)
				types[i] = document.NewRef(id)
			}
		}
		s.Set(&document.Field{
			Key:       key,
			Name:      name,
			Type:      types[i],
			OmitEmpty: omitEmpty,
		})
	}
	return s, nil
}

func (f *jsonFiller) scanArrayType(refOut *document.Document, rootName string) (document.Type, error) {
	types := make([]document.Type, 0)
	for f.More() {
		if t, err := f.scanType(refOut, rootName); err != nil {
			return nil, err
		} else {
			types = append(types, t)
		}
	}
	// Scan end token "]"
	if _, _, err := f.Scan(); err != nil {
		return nil, err
	}
	return document.NewArray(f.deduce(types)), nil
}

func (f *jsonFiller) scanType(refOut *document.Document, rootName string) (document.Type, error) {
	t, _, err := f.Scan()
	if err != nil {
		return nil, err
	}
	switch t {
	case token.LeftBrace:
		return f.scanObjectType(refOut, rootName)
	case token.LeftBracket:
		return f.scanArrayType(refOut, rootName)
	case token.Bool:
		return document.NewBool(), nil
	case token.Int:
		return document.NewInt(), nil
	case token.Float:
		return document.NewFloat(), nil
	case token.String:
		return document.NewString(), nil
	case token.Null:
		return document.NewAny(), nil
	default:
		return nil, fmt.Errorf("unexpected token %s", t)
	}
}

func (f *jsonFiller) normalizeId(parts ...string) string {
	if len(parts) < 1 {
		return ""
	}
	key := strings.TrimPrefix(strings.Join(parts, "."), ".")
	if replaced, ok := f.replacements[key]; ok {
		return replaced
	}
	key = parts[len(parts)-1]
	if replaced, ok := f.replacements[key]; ok {
		return replaced
	}
	return key
}

func (f *jsonFiller) Fill(out *document.Document) error {
	if out == nil {
		return fmt.Errorf("document is nill")
	}
	t, err := f.scanType(out, f.declaredName)
	if err != nil {
		return err
	}
	out.PutDeclaredType(document.NewDeclaredType(f.declaredName, t), true)
	return nil
}

func (f *jsonFiller) UseCaps(a []string) Filler {
	f.camelCtx.UseCaps(a)
	return f
}

func (f *jsonFiller) UseReplacements(a []string) Filler {
	for _, s := range a {
		p := strings.Split(s, ":")
		if len(p) > 1 {
			f.replacements[strings.TrimSpace(p[0])] = strings.TrimSpace(p[len(p)-1])
		}
	}
	return f
}

func NewJsonFiller(data []byte, declaredName string, allCaps ...string) Filler {
	return &jsonFiller{
		Scanner:      scanner.NewFromBytes(data),
		camelCtx:     strcase.NewCamelContext(allCaps),
		replacements: make(map[string]string),
		declaredName: declaredName,
	}
}
