package scanner

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"

	"githouse.ru/macrox/gotygen/pkg/token"
)

type Standard struct {
	*json.Decoder
}

func (s Standard) Scan() (token.Token, string, error) {
	t, err := s.Decoder.Token()
	if err != nil {
		if errors.Is(err, io.EOF) {
			return token.EOF, "", nil
		}
		return token.Illegal, "", err
	}

	switch x := t.(type) {
	case json.Delim:
		switch x {
		case '{':
			return token.LeftBrace, "", nil
		case '}':
			return token.RightBrace, "", nil
		case '[':
			return token.LeftBracket, "", nil
		case ']':
			return token.RightBracket, "", nil
		default:
			return token.Illegal, "", fmt.Errorf("invalid delim %s", x)
		}
	case bool:
		return token.Bool, "", nil
	case json.Number:
		n := x.String()

		if n == "" {
			return token.Float, "", nil
		}
		if !(n[0] == '+' || n[0] == '-' || ('0' <= n[0] && n[0] <= '9')) {
			return token.Float, "", nil
		}
		for i := 1; i < len(n); i++ {
			if !('0' <= n[i] && n[i] <= '9') {
				return token.Float, "", nil
			}
		}

		return token.Int, "", nil
	case string:
		return token.String, "", nil
	case nil:
		return token.Null, "", nil
	default:
		return token.Illegal, "", fmt.Errorf("unexpected type")
	}
}

func NewStandard(s string) Scanner {
	scanner := Standard{
		Decoder: json.NewDecoder(bytes.NewBufferString(s)),
	}
	scanner.UseNumber()
	return scanner
}

func NewStandardFromBytes(data []byte) Scanner {
	scanner := Standard{
		Decoder: json.NewDecoder(bytes.NewBuffer(data)),
	}
	scanner.UseNumber()
	return scanner
}
